(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'checkpoints *mods-alist*))
	  (cons (cons 'checkpoints
		      "This feature, which is documented in CLI
Technical Report 85, assists users with getting
information from the output of a failed proof
attempt.")
		*mods-alist*)
	*mods-alist*)
    nil))

;; At this point I need a pretty clear specification of what I'll checkpoint.

;; Note the following bit of code from Nqthm and Nqthm-1992:

;; (DEFUN ADD-PROCESS-HIST (PROCESS PARENT PARENT-HIST DESCENDANTS HIST-ENTRY)
;;   (IO PROCESS PARENT PARENT-HIST DESCENDANTS HIST-ENTRY)
;;   (CONS (CONS PROCESS (CONS PARENT HIST-ENTRY)) PARENT-HIST))

;; 
;; To keep it simple, we checkpoint each of the following since the last
;; goal pushed for induction (or the proof's start):
;; 
;; goals pushed for induction;
;; 
;; the first non-simplification;
;; 
;; the first execute-process step other than simplification or elimination;
;; 
;; the first generalization.

(defparameter process-print-flag nil)

(defvar total-hist)

;; The following has entries of the form (clause id . dependent-clauses),
;; except that here a ``clause'' can be something like *1.1.
(defvar hist-clause-alist)

;; Here is the list of all printed clause ids.
(defvar all-clause-ids)

(defvar *newly-proved-clause-ids*)

(defvar *saved-random-seed*)

(defun initialize-total-hist ()
  ;;(format (or prove-file t) "~%<<< Beginning checkpointed proof >>>~%")
  (setq hist-clause-alist nil)
  (setq all-clause-ids nil)
  (setq total-hist nil))

(defun remove-proved-clauses-from-hist-clause-alist (proved-clauses a-hist-clause-alist)
  ;; We want to remove all proved clauses, including the given ones
  ;; from a-hist-clause-alist.  This assumes that a-hist-clause-alist respects
  ;; dependencies, in the sense that the dependents of a clause appear "towards
  ;; the front" from that clause.  Note that here a ``clause'' can be something
  ;; like *1.1.

  ;; This has the side effect of setting the global variable *newly-proved-clause-ids*
  ;; to the list of all ids of clauses that have been removed from a-hist-clause-alist.
  (if a-hist-clause-alist
      (let* ((entry (car a-hist-clause-alist))
	     (clause (car entry))
	     (id (cadr entry))
	     (descendants (cddr entry))
	     (new-descendants (set-diff-eq descendants proved-clauses)))
	(if new-descendants
	    (cons (list* clause id new-descendants)
		  (remove-proved-clauses-from-hist-clause-alist
		   proved-clauses
		   (cdr a-hist-clause-alist)))
	  (progn (setq *newly-proved-clause-ids*
		       (cons id *newly-proved-clause-ids*))
		 (remove-proved-clauses-from-hist-clause-alist
		  (cons clause proved-clauses)
		  (cdr a-hist-clause-alist)))))
    nil))

(DEFUN SETUP (FORM CLAUSES LEMMAS)
  (initialize-total-hist)
  (SETQ ORIGTHM FORM)
  (SETQ EXPAND-LST HINTED-EXPANSIONS)
  (SETQ TERMS-TO-BE-IGNORED-BY-REWRITE NIL)
  (SETQ INDUCTION-HYP-TERMS NIL)
  (SETQ INDUCTION-CONCL-TERMS NIL)
  (SETQ ALL-LEMMAS-USED LEMMAS)
  (SETQ STACK NIL)
  (SETQ FNSTACK NIL)
  (SETQ LAST-PRINT-CLAUSES NIL)
  (SETQ TYPE-ALIST NIL)
  (SETQ LITS-THAT-MAY-BE-ASSUMED-FALSE NIL)
  (SETQ CURRENT-LIT 0)
  (SETQ CURRENT-ATM 0)
  (SETQ ANCESTORS NIL)
  (COND (REWRITE-PATH-STK-PTR
         (SETQ REWRITE-PATH-STK-PTR -1)
         (SETQ REWRITE-PATH-FRAME-CNT 0)
         (SETQ REWRITE-PATH-PERSISTENCE-ALIST NIL)))
  (INIT-LEMMA-STACK)
  (INIT-LINEARIZE-ASSUMPTIONS-STACK)
  (SETQ LAST-PRINEVAL-CHAR (QUOTE |.|))
  (RANDOM-INITIALIZATION ORIGTHM)
  (IO (QUOTE SETUP)
      (LIST ORIGTHM)
      NIL CLAUSES LEMMAS))

(DEFUN DEFN-SETUP (EVENT)
  (initialize-total-hist)
  (SETQ ORIGEVENT EVENT)
  (SETQ LAST-PROCESS (QUOTE SETUP))
  (SETQ EXPAND-LST HINTED-EXPANSIONS)
  (SETQ TERMS-TO-BE-IGNORED-BY-REWRITE NIL)
  (SETQ INDUCTION-HYP-TERMS NIL)
  (SETQ INDUCTION-CONCL-TERMS NIL)
  (SETQ STACK NIL)
  (SETQ FNSTACK NIL)
  (SETQ TYPE-ALIST NIL)
  (SETQ LITS-THAT-MAY-BE-ASSUMED-FALSE NIL)
  (SETQ CURRENT-LIT 0)
  (SETQ CURRENT-ATM 0)
  (SETQ ANCESTORS NIL)
  (COND (REWRITE-PATH-STK-PTR
         (SETQ REWRITE-PATH-STK-PTR -1)
         (SETQ REWRITE-PATH-FRAME-CNT 0)
         (SETQ REWRITE-PATH-PERSISTENCE-ALIST NIL)))
  (INIT-LEMMA-STACK)
  (INIT-LINEARIZE-ASSUMPTIONS-STACK)
  (SETQ LAST-PRINEVAL-CHAR (QUOTE |.|))
  (RANDOM-INITIALIZATION ORIGEVENT)
  EVENT)

(defun new-hist-clause-alist (id)
  ;; assumes that we have a checkpointed goal
  (cond
   ((eq process 'induct)
    (cons (list* (car hist-entry) id descendants)
	  hist-clause-alist))
   ((eq process 'store-sent)
    (cons (list parent id (car hist-entry))
	  hist-clause-alist))
   (descendants
    (cons (list* parent id descendants)
	  hist-clause-alist))
   ((not (eq process 'finished))
    (remove-proved-clauses-from-hist-clause-alist
     (setq *newly-proved-clause-ids* (list parent))
     hist-clause-alist))))

(defun io2 ()
  (let* ((checkpoint-info (make-checkpoint process))
	 (id (length total-hist))
	 (*newly-proved-clause-ids* nil))
    (setq total-hist
	  (cons (list (cons PROCESS checkpoint-info) PARENT PARENT-HIST DESCENDANTS HIST-ENTRY)
		total-hist))
    (setq hist-clause-alist
	  (new-hist-clause-alist id))
    (print-checkpoint-info process checkpoint-info)
    (print-removed-clause-ids *newly-proved-clause-ids*)
    (io1)))

(setq io-fn 'io2)

(DEFUN RANDOM-INITIALIZATION (EVENT)
  (setq *saved-random-seed*
	(SETQ *RANDOM-SEED* (CONS-COUNT EVENT))))

;; We make a checkpoint when a goal is pushed for induction,
;; when we do our first non-simplification, and when we do our
;; first non-simplification&non-elimination.  The result is
;; to add a "label" to the process component of a history element.

(defparameter non-simplification-execute-processes
  (QUOTE (;;  SIMPLIFY-CLAUSE SETTLED-DOWN-CLAUSE
	  FERTILIZE-CLAUSE
	  ELIMINATE-DESTRUCTORS-CLAUSE GENERALIZE-CLAUSE
	  ELIMINATE-IRRELEVANCE-CLAUSE STORE-SENT)))

(defparameter non-simplification-elimination-execute-processes
  (QUOTE (;;  SIMPLIFY-CLAUSE SETTLED-DOWN-CLAUSE
	  FERTILIZE-CLAUSE
	  ;;  ELIMINATE-DESTRUCTORS-CLAUSE
	  GENERALIZE-CLAUSE
	  ELIMINATE-IRRELEVANCE-CLAUSE STORE-SENT)))

(defparameter io-execute-processes
  (QUOTE (SIMPLIFY-CLAUSE
	  ;; SETTLED-DOWN-CLAUSE
	  FERTILIZE-CLAUSE
	  ELIMINATE-DESTRUCTORS-CLAUSE
	  GENERALIZE-CLAUSE
	  ELIMINATE-IRRELEVANCE-CLAUSE STORE-SENT)))

(defun processes-not-seen-p (history processes)
  (let (process)
    (if (and history
	     (not (eq (setq process (caaar history)) 'store-sent)))
	(and (not (member-eq process processes))
	     (processes-not-seen-p (cdr history) processes))
      t)))

(defun some-process-seen-p (history)
  (if history
      (or (member-eq (caaar history) io-execute-processes)
	  (some-process-seen-p (cdr history)))
    nil))

(defun current-store-sent-level ()
  ;; could make this more efficient by storing the level, most likely
  (current-store-sent-level-rec total-hist))

(defun current-store-sent-level-rec (hist)
  (if hist
      (if (eq (caaar hist) 'store-sent)
	  (1+ (current-store-sent-level-rec (cdr hist)))
	(current-store-sent-level-rec (cdr hist)))
    0))

(defvar checkpoint-priority-alist
  ;; Do not modify this alist, except possibly to shuffle the numbers.
  ;; FERTILIZE-CLAUSE and ELIMINATE-IRRELEVANCE-CLAUSE must both be
  ;; assigned the same value, and other than that all processes must
  ;; be assigned different values, and all between 0 and 4, where 0
  ;; is the most important.  The only place we rely on this specification
  ;; is in the emacs code, where there is a maximum of 4 in the variable
  ;; max-priority, and where only the first priority at a given store-sent-level
  ;; is recognized.
  '((ELIMINATE-DESTRUCTORS-CLAUSE . 0)
    (FERTILIZE-CLAUSE . 3)
    (ELIMINATE-IRRELEVANCE-CLAUSE . 3)
    (GENERALIZE-CLAUSE . 1)
    (STORE-SENT . 2)
    (INDUCT . 4)))

(defun checkpoint-priority (process)
  (or (cdr (assoc-eq process checkpoint-priority-alist))
      (er hard (process) |Attempted| |to| |assign| |a|
	  |priority| |to| |process| (!ppr process (quote |.|)))))

(defun checkpoint ()
  (let ((id (length total-hist)))
    (setq all-clause-ids (cons id all-clause-ids))
    (list (current-store-sent-level)
	  (checkpoint-priority process)
	  id				;unique identifier
	  (and process-print-flag process))))

(defun make-checkpoint (process)
  ;; note that settled-down-clause and store-sent
  ;; should not be among current-processes-seen
  (if (and (member-eq process execute-processes)
	   (some-process-seen-p total-hist))
      (case process
	    (ELIMINATE-DESTRUCTORS-CLAUSE
	     ;; make checkpoint if this is the first non-simplification
	     (when (processes-not-seen-p total-hist
					 non-simplification-execute-processes)
		   (checkpoint)))
	    ((SIMPLIFY-CLAUSE SETTLED-DOWN-CLAUSE)
	     nil)
	    ((FERTILIZE-CLAUSE ELIMINATE-IRRELEVANCE-CLAUSE)
	     (when (processes-not-seen-p total-hist
					 non-simplification-elimination-execute-processes)
		   (checkpoint)))
	    (GENERALIZE-CLAUSE
	     ;; always checkpoint a generalization
	     (checkpoint))
	    ;; make checkpoint if this is the first non-simplification-elimination
	    (otherwise			;STORE-SENT
	     (checkpoint)))
    (if (eq process 'induct)
	(checkpoint)
      nil)))

(defun process-print-name (process)
  (or (cdr (assoc-eq process
		     '((simplify-clause . simplify)
		       (fertilize-clause . cross-fertilize)
		       (eliminate-destructors-clause . eliminate-destructors)
		       (generalize-clause . generalize)
		       (eliminate-irrelevance-clause . eliminate-irrelevance))))
      process))

(defun print-checkpoint-info (process ci)
  (when ci
	(format (or prove-file t)
		"~%~%!!CHECKPOINT LEVEL ~s; PRIORITY ~s; ID ~s"
		(car ci) (cadr ci)
		(caddr ci)		;unique identifier
		)
	(when process-print-flag
	      (format (or prove-file t)
		      "~&~s"
		      (process-print-name (cadddr ci))))))

(defun print-removed-clause-ids (ids)
  (let ((printed-ids (intersection ids all-clause-ids)))
    (when printed-ids
	  (format (or prove-file t) "~&~%!!NOTE PROVED CLAUSES ~s" printed-ids))))

(defmacro pf (form &optional (filename "proof.out"))
  `(with-open-file (prove-file ,filename :direction :output)
		   ,form))

(defmacro checkpoint-y-or-n-p (msg var yes-ans no-ans &rest args)
  `(progn
     ,(list* 'format t msg args)
     (let ((ans (read t)))
       (cond
	((or (eq ans 'y)
	     (eq ans 'yes))
	 (setq ,var ,yes-ans))
	((or (eq ans 'n)
	     (eq ans 'no))
	 (setq ,var ,no-ans))
	((eq ans 'a)
	 (throw 'checkpoint-options 'abort))
	((eq ans 's)
	 (format t "~&  No change.~%"))
	(t (format t "~&  Response unknown... no change.~%"))))))

(defun checkpoint-options ()
  (catch 'checkpoint-options
    (format t "~&Please answer each question with either Y (yes), N (no), A (abort queries), or~%~
             S (Skip) or any other character (if you don't want a change).~&~%~
             For a similar utility in emacs, do control-t c.~%~%")
    (checkpoint-y-or-n-p "Do you want checkpoint-printing on (currently ~a)?  "
			 io-fn 'io2 'io1
			 (if (eq io-fn 'io2) "off" "on"))
    (checkpoint-y-or-n-p "Do you want process names printed with checkpoints (currently ~a)?  "
			 process-print-flag t nil
			 (if process-print-flag "on" "off"))
    (let (ans)
      (and (checkpoint-y-or-n-p "Would you like information about changing priorities?  "
				ans t nil)
	   (format t "~%To change priorities, (setq checkpoint-priority-alist <new-alist>)~%~
                      where <new-alist> looks like the current value,~%~%   ~s.~%~%~
                      Note however that all priorities should be between 0 and 4 (or you'll~%~
                      have to modify emacs), and that only the first checkpoint at a given~%~
                      priority will be read by emacs.~%~%"
		   checkpoint-priority-alist)))))
