#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

;; Functions in this file where arity-alist has been extended:
;; DEFN-MUTUAL-FORMS (for DEFN-MUTUAL-BODY), TRANSLATE-BODIES,
;; and CHK-DEFN-MUTUAL-HINTS.  Since none of these is called inside
;; any of the others, this should all be OK.

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'mutual-recursion *mods-alist*))
	  (cons (cons 'mutual-recursion
		      "A facility for making mutually recursive
definitions in Nqthm.  See \"mutual.doc\" for documentation.")
		*mods-alist*)
	*mods-alist*)
    nil))

;;;;;;;;;   PART 1:  Defining the primitive DEFN-MUTUAL   ;;;;;;;;;

;; We want to implement:

;(defn-mutual
;  ((foo foo-args foo-body)
;   ...
;   (bar bar-args bar-body))
;  &optional hints interactive-hint)

; where each hint is of the form, (relation (foo foo-measure) ... (bar bar-measure)).

(defvar *mutual-flg* 'mutual-flg)

(defvar *prefix-for-mutual* 'mutual)

(defvar *remaining-mutual-events*)

(defun defn-mutual-body1 (newname arglist defns term)
  ;; returns new version of term in which each call of any function defined in defns is replaced by a
  ;; corresponding call of newname with the appropriate flag
  (let (triple)
    (cond
     ((or (variablep term) (fquotep term))
      term)
     ((setq triple (assoc-eq (ffn-symb term) defns))
      (cons-term newname
		 (cons (list 'quote (ffn-symb term))
		       (iterate for arg in arglist
				with index and args = (cadr triple)
				collect
				(cond ((setq index (position arg args))
				       (defn-mutual-body1 newname arglist defns
					 (fargn term (1+ index))))
				      (t true))))))
     (t
      (cons-term (ffn-symb term)
		 (iterate for arg in (fargs term)
			  collect (defn-mutual-body1 newname arglist defns arg)))))))

(defun defn-mutual-body (newname arglist all-defns remaining-defns)
  ;; assumes that arity-alist has been set to include entries for the new names
  (let ((next-body (translate (caddar remaining-defns))))
    (if (null (cdr remaining-defns))
	(defn-mutual-body1 newname arglist all-defns next-body)
      `(if (equal ,*mutual-flg* ',(caar remaining-defns))
	   ,(defn-mutual-body1 newname arglist all-defns next-body)
	 ,(defn-mutual-body newname arglist all-defns (cdr remaining-defns))))))

(defun mutual-name (defns)
  (pack (cons *prefix-for-mutual*
	      (iterate for triple in defns append (list '- (car triple))))))

(defun mutual-arglist (defns)
  (unstable-sort
   (remove-duplicates
    (iterate for x in (mapcar 'cadr defns)
	     append x))
   (function alphorder)))

(defun defn-mutual-forms (defns &rest rest
			   &aux arglist newname body)
  (iterate for arg in defns do (chk-new-name (car arg) nil))
  (setq arglist (mutual-arglist defns))
  (setq newname (mutual-name defns))
  (let ((arity-alist
	 (append (iterate for triple in defns collect
			  (cons (car triple) (length (cadr triple))))
		 arity-alist)))
    (setq body
	  (defn-mutual-body newname arglist defns defns)))
  (cons `(defn ,newname ,(cons *mutual-flg* arglist) ,(untranslate body) ,@rest)
	(setq *remaining-mutual-events*
	      (iterate for arg in defns
		       collect `(defn ,(car arg) ,(cadr arg)
				  ;; Notice that UNTRANSLATE doesn't object to the new name
				  ,(untranslate
				    (defn-mutual-body1 newname arglist defns
				      (cons-term (car arg) (cadr arg)))))))))

(defun do-events-recursive (ev)
  (let (undone-events)
    (do-events ev)))

(defmacro defn-mutual (defns &optional hints instructions)
  (if instructions
      `(do-events-recursive (defn-mutual-forms ',defns ',hints ',instructions))
    (if hints
	`(do-events-recursive (defn-mutual-forms ',defns ',hints))
      `(do-events-recursive (defn-mutual-forms ',defns)))))

(defun finish-mutual ()
  (do-events-recursive *remaining-mutual-events*))

;; Notice that the handling of interactive mutually recursive definitions should
;; compile OK even if VERIFY-DEFN hasn't been defined, since it's merely "data"
;; here (to be fed to EVAL).

(defun verify-defn-eval (args)
  (eval `(verify-defn ,@args)))

(defmacro verify-defn-mutual (defns &optional hints)
  `(progn
     (prineval
      (pquote (progn CR |Entering| |an| |interactive| DEFN |event| |.|
		     |Submit| |(FINISH-MUTUAL)| |to| |then| |run| |the|
		     |remaining| |events| |defining| |the| |individual|
		     |functions| |,| (!ppr-list names '|.|) CR CR))
      (list (cons 'names
		  (iterate for defn in ',defns
			   collect (car defn))))
      0 prove-file) 
     (verify-defn-eval (cdar (defn-mutual-forms ',defns ',hints)))))

;;;;;;;;; Part 2:  Processing sequences of DEFNs which may    ;;;;;;;;;
;;;;;;;;;                  be mutually recursive              ;;;;;;;;;

;; Here I wish to do a bit of extra analysis to DEFN-MUTUAL so that
;; only necessary mutual recursion is tracked.  The idea is to take
;; the given DEFN forms, translate all their bodies, figure out all
;; the function names for each body, and form a resulting list of DEFN
;; and DEFN-MUTUAL events as a result.

;; Given a transitive relation, we'd like to find a preorder which
;; respects that relation in the strong sense that equivalence is the
;; same in the given relation as it is in the preorder.
;; Operationally, what this means to us is that two functions depend
;; on each other if and only if they are ultimately defined in the
;; same DEFN-MUTUAL event.

;; So, what we'll do is the following.

;; 1. TRANSLATE-DEFNS translates the bodies of the given defns (assuming
;;    an arity-alist where each given function "exists").
;; 2. DEFN-DEPENDENCIES-ALIST creates the appropriate dependency relation
;;    from the translated defns.
;; 3. MUTUAL-TRANSITIVE-CLOSURE takes a binary relation, represented as a alist,
;;    and returns a new alist which represents its transitive closure.
;; 4. SQUASH-TRANSITIVE-RELATION takes a transitive relation, represented
;;    as an alist, and returns a list representing the preorder with the
;;    property described above (of preserving the relation without changing
;;    the notion of equivalence).

(defun translate-bodies (defns)
  ;; Returns a list of translated defns, when given a list of
  ;; defns (all this sans DEFN).
  (let ((arity-alist
	 (append (iterate for defn in defns
			  collect (cons (car defn) (length (cadr defn))))
		 arity-alist)))
    (iterate for defn in defns
	     with temp
	     collect
	     (progn (setq temp (error1-set (translate (caddr defn))))
		    (if temp
			(car temp)
		      (er soft ((name (car defn)))
			  |Translation| |error| |occurred| |in| |definition| |of|
			  (!ppr name (quote |.|))))))))

(defun defn-dependencies-alist (names bodies)
  ;; assumes that the bodies have been translated
  (iterate for name in names
	   as body in bodies
	   collect (cons name
			 (intersection-eq names (all-fnnames body)))))

(defun mutual-transitive-close1 (alist)
  ;; Here alist is a list of triples (atom . new . old), where
  ;; we assume that all dependents of old occur in old or in new.
  (iterate for x in alist
	   collect (cons (car x)
			 (cons (remove-duplicates
				(set-diff
				 (iterate for y in (cadr x)
					  append (cddr (assoc-eq y alist)))
				 (cddr x)))
			       (union-eq (cadr x) (cddr x))))))

(defun mutual-transitive-closure (alist)
  (let ((a (iterate for x in alist
		    collect (cons (car x) (cons (cdr x) (cdr x))))))
    (iterate with old-a
	     while (not (equal old-a a))
	     do (progn (setq old-a a)
		       (setq a (mutual-transitive-close1 a)))
	     finally (return (iterate for x in a
				      collect (cons (car x) (cddr x)))))))

(defun insert-for-squash (elt list alist)
  ;; Here alist represents a given transitive relation and
  ;; list is a list of non-empty lists representing a preorder.  This
  ;; returns the result of inserting elt into an element of
  ;; list, so that the resulting preorder continues to respect
  ;; the given alist, i.e. to extend the given transitive relation
  ;; (at least as far as list goes so far) and to identify as
  ;; equivalent two elements if and only if they were equivalent
  ;; according to the given alist.
  (cond
   ((null list)
    (list (list elt)))
   ((member-eq elt (cdr (assoc-eq (caar list) alist)))
    ;; so, the first element of list (viewed as part of an equivalence
    ;; class of alist) depends on elt, and hence elt should be listed
    ;; first unless the dependency goes the other way too
    (if (member-eq (caar list) (cdr (assoc-eq elt alist)))
	(cons (cons elt (car list)) (cdr list))
      (cons (list elt) list)))
   (t
    ;; so, the given elt doesn't include the first element(s) of list
    ;; among its dependents
    (cons (car list)
	  (insert-for-squash elt (cdr list) alist)))))

(defun squash-transitive-relation (alist)
  (iterate for x in alist
	   with list
	   do (setq list (insert-for-squash (car x) list alist))
	   finally (return list)))

(defun chk-defn-mutual-hints (defns &aux relation rest temp first-name-having-hint measure)
  ;; Here each member of DEFNS starts with the name, i.e. the DEFN header is omitted
  ;; Causes an error if things aren't OK.
  (let ((arity-alist
	 (append (iterate for defn in defns
			  collect (cons (car defn) (length (cadr defn))))
		 arity-alist)))
    (or
     ;; first case:  all hints are NIL -- if not, set RELATION
     (iterate for x in defns
	      always (or (null (cdddr x))
			 (cond ((match (cadddr x)
				       (cons (list relation measure) &))
				(setq first-name-having-hint (car x))
				nil)
			       (t (er soft ((hints (cadddr x)))
				      |Illegal| |hints| |,| (!ppr hints (quote |.|)))))))
     (iterate for x in defns
	      always (cond ((match (cadddr x)
				   (cons (list temp measure) rest))
			    (cond ((not (subsetp-eq (all-vars (translate measure))
						   (cadr x)))
				   (er soft (measure (name (car x)))
				       |Sorry| |,| |but| |the| |measure| |term| (!ppr measure nil)
				       |in| |the| |hint| |for| |the| |definition| |of| (!ppr name nil)
				       |has| |variables| |not| |included| |in| |the| |list| |of|
				       |formal| |parameters| |for| |that| |definition| |.|))
				  ((not (eq temp relation))
				   (er soft (relation first-name-having-hint temp (name (car x)))
				       |The| |relation| |must| |be| |the| |same| |in| |each|
				       |hint| |to| |a| DEFN |in| |a| |set| |of| |mutually|
				       |recursive| |definition| |,| |but| |the| |relation|
				       (!ppr relation nil) |from| |the| |definition| |of|
				       (!ppr first-name-having-hint nil) |differs| |from| |the| |relation|
				       (!ppr temp nil) |from| |the| |definition| |of|
				       (!ppr name (quote |.|))))
				  (rest
				   (er warning ((name (car x)) rest)
				       |The| |extra| |hints| |,| (!ppr rest (quote |,|))
				       |in| |the| |definition| |of| (!ppr name (quote |,|)) |are|
				       |being| |ignored| |.|)
				   t)
				  (t t)))
			   (t (er soft ((name (car x)) first-name-having-hint)
				  |The| |definition| |of| (!ppr first-name-having-hint nil)
				  |has| |a| |hint| |but|
				  |the| |definition| |of| (!ppr name nil) |does| |not| |.|
				  |This| |is| |illegal| |for| |mutual| |recursion| |.|)))))))

(defun defn-mutual-measure (newname arglist relation defns)
  (let (name measure)
    (match (car defns)
	   (list name & & (cons (list & measure) &)))
    (setq measure
	  (defn-mutual-body1 newname arglist defns measure))
    (if (null (cdr defns))
	measure
      `(if (equal ,*mutual-flg* ',name)
	   ,measure
	 ,(defn-mutual-measure newname arglist relation (cdr defns))))))

(defun make-defn-mutual (defns &aux (relation (caaar (cdddr (car defns)))))
  ;; Takes a list of DEFN events (without their DEFN headers) and
  ;; returns a DEFN-MUTUAL event which glues together their hints,
  ;; assuming that each of their hints is of the form
  ;; ((<relation> <term>)) where <relation> is the same each time.
  (chk-defn-mutual-hints defns)
  `(defn-mutual
     ,(iterate for x in defns
	       collect (list (car x) (cadr x) (caddr x)))
     
     ,@(if (cdddr (car defns))
	   (list (list (list relation
			     (defn-mutual-measure (mutual-name defns) (mutual-arglist defns)
			       relation defns))))
	 nil)))

(defun defn-and-defn-mutual-events (defns)
  ;; Takes a list of DEFN events (without their DEFN headers) and
  ;; returns a list of DEFN and DEFN-MUTUAL events which is "nice",
  ;; i.e. avoids unnecessary dependencies.
  (iterate for namelist in
	   (squash-transitive-relation
	    (mutual-transitive-closure
	     (defn-dependencies-alist
	       (mapcar #'car defns)
	       (translate-bodies defns))))
	   collect
	   (if (null (cdr namelist))
	       (cons 'defn (assoc-eq (car namelist) defns))
	     (make-defn-mutual
	      (iterate for name in namelist
		       collect (assoc-eq name defns))))))

(defun do-mutual (defns)
  ;; Here the DEFN events DO have their DEFN headers.  Notice that by
  ;; using DO-EVENTS-RECURSIVE, we don't get access to the undone events
  ;; when we're through, but on the other hand we can use this form
  ;; within a DO-EVENTS without harm.
  (iterate for defn in defns
	   always (or (match defn (cons (quote defn) &))
		      (er soft (defn) |The| |form| DO-MUTUAL |is| |only|
			  |applicable| |to| |sequences| |of| DEFN |forms| |,|
			  |but| (!ppr defn nil) |is| |not| |.|)))
  (do-events-recursive (defn-and-defn-mutual-events (mapcar #'cdr defns))))
