#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'purify *mods-alist*))
	  (cons (cons 'purify
		      "A facility for creating event files from
libraries, even when those libraries were created using
various hacks rather than a `pure' Nqthm or Pc-Nqthm.
See \"purify.doc\" for documentation.")
		*mods-alist*)
	*mods-alist*)
    nil))

;; The purpose of this utility is to allow one to take a completed
;; proof effort using (pc-)nqthm and "purify" it a sense I'll now
;; describe.

;; Nqthm users commonly write macros that generate events.  Also,
;; nqthm users at CLI tend to use the "saved_pc-nqthm" core image,
;; which includes local modifications to pc-nqthm.  In some cases
;; there's really no use of pc-nqthm at all beyond what's in nqthm,
;; except perhaps for theories; hence I've also created a file
;; "deftheory.lisp" that contains the disabledp patch and deftheory,
;; but only these.  (Probably I could have left out the disabledp
;; patch, but I didn't want to think about the ramifications that
;; would have for deftheory.)

;; At any rate, once one chooses a core image that corresponds to
;; one of the the publicly released systems -- i.e. either nqthm
;; or pure-pc-nqthm -- one might like to specify some additional
;; forms to execute (perhaps to load files like "deftheory" or
;; "fast-clausifier") and also a library to replay.  The effect should
;; be as illustrated below.

;; 1. The library is used to create an events file.

;; 2. Nqthm or pure-pc-nqthm is started up, with the indicated forms
;;    executed.

;; 3. A proveall is run using the events file created in the first step.

;; A sample transcript occurs at the end of the file.

(defmacro library-to-events (filename
			     &optional event-filename events-list-name
			     &rest forms)
  ;; e.g. (library-to-events "foo" "foo.events") should write to
  ;; "foo.events" a list of events that could be run in order to create
  ;; the library "foo", i.e. "foo.lib" and "foo.lisp".  This should work
  ;; even if the filenames have directory prefixes.
  (setq event-filename
	(or event-filename
	    (concatenate 'string filename "-replay.events")))
  (setq events-list-name
	(or events-list-name
	    (concatenate 'string filename "-replay")))
  `(progn
     (note-lib ,filename)
     (database-to-events ,event-filename ,events-list-name ',forms)))

(defun database-to-events (event-filename events-list-name forms)
  ;; Writes out a list of events to event-filename that could be run
  ;; in order to create the current prover state.
  (with-open-file
   (str event-filename :direction :output)
   (format t "~%Creating events file ~s...." event-filename)
   (when forms
	 (format str "~s~%~%" (if (cdr forms) (cons 'progn forms) (car forms)) str))
   (format str "(PROVEALL ~s '(~%" events-list-name)
   (iterpri str)
   (iterate for name in (reverse chronology)
	    do
	    (progn
	      (ppr (untranslate-event (get name (quote event)))
		   str)
	      (iterpri str)
	      (iterpri str)))
   (format str "~%))~%")
   (format t " done.~%")))

#| Sample transcript.  It assumes that we used pc-nqthm on the following
events to create library "/usr/home/kaufmann/foo", i.e. "foo.lib" and "foo.lisp"
in that same directory.

(do-events '(

(prove-lemma times-comm (rewrite)
  (equal (times x z) (times z x)))

(defn blob (x)
  (let ((y (plus x x)))
    (times y 3)))

(prove-lemma times-add1 (rewrite)
  (equal (times (add1 x) y)
	 (plus y (times x y))))

(prove-lemma blob-val nil
  (equal (blob x)
	 (cond ((zerop x) 0)
	       (t (times 6 x)))))
))

.... and, here's what we do to create a pure nqthm events file
"foo-replay.events":

client12:kaufmann[77]% pc-nqthm
AKCL (Austin Kyoto Common Lisp)  Version(1.527) Mon Jan  7 12:51:51 CST 1991
Contains Enhancements by W. Schelter

Nqthm, with functional instantiation.
Initialized with (BOOT-STRAP NQTHM) on January 7, 1991  14:22:37.
>(load "purify.o")
Loading purify.o
start address -T 30b800 Finished loading purify.o
1864

>(library-to-events "/usr/home/kaufmann/foo" nil nil (print "hi") (print "there"))
To print patch messages, (PRINT-NQTHM-PATCH-DISCLAIMER)
Loading /usr/home/kaufmann/foo.lib
Finished loading /usr/home/kaufmann/foo.lib
Loading /usr/home/kaufmann/foo.lisp
Finished loading /usr/home/kaufmann/foo.lisp

Creating events file "/usr/home/kaufmann/foo-replay.events".... done.
NIL

>(bye)
Bye.
client12:kaufmann[78]% nqthm
AKCL (Austin Kyoto Common Lisp)  Version(1.527) Mon Jan  7 12:51:51 CST 1991
Contains Enhancements by W. Schelter

Nqthm, with functional instantiation.
Initialized with (BOOT-STRAP NQTHM) on January 7, 1991  14:22:37.
>(load "/home/kaufmann/foo-replay.events")
Loading /home/kaufmann/foo-replay.events

"hi" 
"there" 
NQTHM,TIMES-COMM,BLOB,TIMES-ADD1,BLOB-VAL,
Total Statistics:
[ 6.6 18.0 1.8 ]

(MAKE-LIB /usr/home/kaufmann/foo-replay)
Finished loading /home/kaufmann/foo-replay.events
T

>

|#
