#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'queries *mods-alist*))
	  (cons (cons 'queries
		      "Some additional query capabilities.
See \"queries.doc\" for documentation.")
		*mods-alist*)
	*mods-alist*)
    nil))

;; Courtesy of Bill Bevier

(defun axioms ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'add-axiom)))
		 chronology))

(defun defns ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'defn)))
		 chronology))

(defun shells ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'add-shell)))
		 chronology))

(defun lemmas ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'prove-lemma)))
		 chronology))

(defun toggles ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'toggle)))
		 chronology))

(defun deftheories ()
  (remove-if-not (function (lambda (x) (eq (car (get x 'event)) 'deftheory)))
		 chronology))

(defun matching-rules (term)
  ;; Return a list of all lemmas which match TERM, 
  ;; in the sense that RELIEVE-HYPS may be called
  ;; from REWRITE-WITH-LEMMAS on each of these lemmas
  ;; (if enabled) when simplifying TERM.
  (let (unify-subst inst-hyp)
    (cond ((variablep term) nil)
	  ((fquotep term) nil)
	  (t (iterate for lemma in (get (ffn-symb term) (quote lemmas))
		      appending (let ((event-pair (match-lemma term lemma)))
				  (if event-pair (list (car event-pair)) nil)))))))

(defun match-lemma (term lemma)
  ;; This function returns a pair [LEMMA UNIFY-SUBST] if RELIEVE-HYPS
  ;; would be called on this lemma when simplifying TERM. In fact, this
  ;; code is just the body of the big loop in REWRITE-WITH-LEMMAS,
  ;; modified to return LEMMA wherever a call to RELIEVE-HYPS occurs.
  ;; There is another path to RELIEVE-HYPS in REWRITE-WITH-LEMMAS which goes
  ;; through the linear arithmetic package. This patch isn't accounted for.
  ;; BEVIER
  ;; 
  (let
      ;; These variable settings permit MATCH-LEMMA to win based purely
      ;; on syntactic issues. A rule may in fact not be fired at simplify
      ;; time due to the dynamic state of the theorem prover.
      ((objective (quote ?)) 
       (id-iff (quote iff))
       (fnstack nil))
    (cond ((eq (ffn-symb (access rewrite-rule concl lemma))
	       (quote not))
	   (cond ((and (or (null (access rewrite-rule hyps lemma))
			   (not (eq objective (quote true))))
		       (one-way-unify
			(fargn (access rewrite-rule
				       concl lemma)
			       1)
			term))
		  (list (get (access rewrite-rule name lemma) 'event)
			unify-subst))
		 (t nil)))
	  ((eq (ffn-symb (access rewrite-rule concl lemma))
	       (quote equal))
	   (cond ((and
		   (or (null (access rewrite-rule hyps lemma))
		       (not (eq objective (quote true)))
		       (not (equal (fargn (access rewrite-rule
						  concl lemma)
					  2)
				   false)))
		   (or (not (member-eq (ffn-symb term) fnstack))
		       (not (fnnamep (ffn-symb term)
				     (fargn (access rewrite-rule
						    concl lemma)
					    2))))
		   (one-way-unify
		    (fargn (access
			    rewrite-rule concl lemma)
			   1)
		    term)
		   ;; (progn (setq temp commuted-equality-flg) t)
		   (iterate for pair
			    in (access rewrite-rule
				       loop-stopper lemma)
			    never (term-order
				   (cdr (assoc-eq (car pair)
						  unify-subst))
				   (cdr (assoc-eq (cdr pair)
						  unify-subst)))))
		  (list (get (access rewrite-rule name lemma) 'event)
			unify-subst))
		 ((and (or (null (access rewrite-rule hyps lemma))
			   (not (eq objective (quote false))))
		       (eq (ffn-symb term)
			   (quote equal))
		       (one-way-unify (access rewrite-rule concl
					      lemma)
				      term))
		  (list (get (access rewrite-rule name lemma) 'event)
			unify-subst))
		 (t nil)))
	  ((and (eq (ffn-symb (access rewrite-rule concl lemma))
		    (quote iff))
		(eq id-iff (quote iff)))
	   (cond ((and
		   (or (null (access rewrite-rule hyps lemma))
		       (not (eq objective (quote true)))
		       (not (equal (fargn (access rewrite-rule
						  concl lemma)
					  2)
				   false)))
		   (or (not (member-eq (ffn-symb term) fnstack))
		       (not (fnnamep (ffn-symb term)
				     (fargn (access rewrite-rule
						    concl lemma)
					    2))))
		   (one-way-unify
		    (fargn (access
			    rewrite-rule concl lemma)
			   1)
		    term)
		   (iterate for pair
			    in (access rewrite-rule
				       loop-stopper lemma)
			    never (term-order
				   (cdr (assoc-eq (car pair)
						  unify-subst))
				   (cdr (assoc-eq (cdr pair)
						  unify-subst)))))
		  (list (get (access rewrite-rule name lemma) 'event)
			unify-subst))
		 (t nil)))
	  ((and (or (null (access rewrite-rule hyps lemma))
		    (not (eq objective (quote false))))
		(or (eq id-iff (quote iff))
		    (boolean term))
		(one-way-unify (access rewrite-rule concl lemma)
			       term))
	   (list (get (access rewrite-rule name lemma) 'event)
		 unify-subst))
	  (t nil))))

(defun events-apropos (stringlist &optional event-type)  
  ;; Return a list of all events whose names contain every element of
  ;; STRINGLIST as a substring.  
  ;; e.g. (EVENTS-APROPOS '(FOO BAR)) returns a list of the names of
  ;;      all events which contain both "FOO" and "BAR" as substrings.
  ;;
  ;;      (EVENTS-APROPOS 'FOO) returns a list of the names of all
  ;;      events which contain "FOO" as a substring. 
  ;;
  ;; The search can be restricted to a given type of event with the
  ;; EVENT-TYPE argument. 
  ;; Bevier, 9/7/86.
  (setq stringlist
	(mapcar (function (lambda (x) 
			    (if (symbolp x)
				(string x)
				(if (stringp x)
				    (string-upcase x)
				    (error "~%Invalid argument: ~a" x)))))
		(if (listp stringlist)
		    stringlist
		    (list stringlist))))

  (if (null event-type) ;; testing for presence of EVENT-TYPE before
                        ;; search is faster.

      (remove-if-not
       (function (lambda (event-name)
		   (iterate for string in stringlist
			    always (search string 
					   (string event-name)))))
       chronology)

      (remove-if-not
       (function (lambda (event-name)
		   (and (eq (car (get event-name 'event)) event-type)
			(iterate for string in stringlist
				 always (search string
						(string event-name))))))
       chronology)))

;; The following is taken literally from version 1.1 of pc-nqthm
(DEFUN MAKE-NEW-IND-INSTANCE (TERM TESTS-AND-CASES FORMALS VARS)
  (LET ((TESTS (ACCESS TESTS-AND-CASES TESTS TESTS-AND-CASES))
	(CASES (ACCESS TESTS-AND-CASES CASES TESTS-AND-CASES))
	HYP)
    (SETQ TESTS
	  (SUB-PAIR-VAR-LST FORMALS VARS TESTS))
    (SETQ CASES
	  (ITERATE FOR CASE IN CASES
		COLLECT
		  (SUB-PAIR-VAR-LST FORMALS VARS CASE)))
    (SETQ HYP
	  (DUMB-CONJOIN
	    (APPEND TESTS
		    (ITERATE FOR CASE IN CASES
			  COLLECT
			    (SUB-PAIR-VAR VARS CASE TERM)))))
    (FCONS-TERM* (QUOTE IMPLIES) HYP TERM)))

(DEFUN INDUCTION-GOALS (HINT TERM)
  (if (fboundp 'interactively-rewrite)
      (COND
       ((AND				; Pretty much lifted from the INDUCT case of
					; the BMTP definition for CHK-ACCEPTABLE-HINTS
	 (EQUAL HINT (TRANSLATE HINT))
	 (NVARIABLEP HINT)
	 (NOT (FQUOTEP HINT))
	 (GET (FFN-SYMB HINT) (QUOTE JUSTIFICATIONS))
	 (GET (FFN-SYMB HINT) (QUOTE INDUCTION-MACHINE))
	 (GET (FFN-SYMB HINT) (QUOTE SDEFN))
	 (ITERATE FOR V IN (FARGS HINT) ALWAYS (VARIABLEP V))
	 (NO-DUPLICATESP (FARGS HINT)))
	(ITERATE FOR TAC IN (GET (FFN-SYMB HINT)
				 (QUOTE INDUCTION-MACHINE))
		 WITH FORMALS =
		 (CADR (GET (FFN-SYMB HINT) (QUOTE SDEFN)))
		 AND VARS = (FARGS HINT)
		 COLLECT
		 (MAKE-NEW-IND-INSTANCE
		  (TRANSLATE TERM) TAC FORMALS VARS)))
       (T (ER WARNING ((H (QUOTE (INDUCT (|fn| |v1| |...| |vn|)))))
		 |**NO| |CHANGE**| |--| |The| |first| |argument| |of| INDUCTION-GOALS |must| |have|
		 |the| |form| (!PPR H NIL) |where|
		 |fn| |is| |a| |recursively| |defined| |function| |and|
		 |the| |vi| |are| |distinct| |variables| |of| |a| |number|
		 |matching| |the| |arity| |of| |fn| |.|)))
    (format t "~&SORRY, but INDUCTION-GOALS doesn't~
                     currently~%work without the proof-checker loaded.~%")))

(defun deadwood ()
  (iterate for x in chronology
	   when (and (not (eq (car (get x 'event)) 'toggle))
		     (null (data-base 'immediate-dependents x)))
	   collect x))

(defun more-deadwood (&optional (namelist (deadwood)))
  (iterate for x in chronology
	   when (and (not (eq (car (get x 'event)) 'toggle))
		     (not (member-eq x namelist))
		     (subsetp-eq (data-base 'immediate-dependents x) namelist))
	   collect x))
