#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

; This is a utility for introducing abbreviations into Nqthm output.
; Here is how to use this facility.

; ADD AN ABBREVIATION:

; Examples:  (abb '(1 2 3 4 5 6 7 8 9 10) @10-numbers)
;            (abb (append x y) app-x-y)

; (abb term val)

; ABB causes the given term (which may be presented in ordinary
; user-level syntax) to be printed by Nqthm as val whenever it is
; encountered.  In the example above, '(1 2 3 4 5 6 7 8 9 10) would be
; printed as @10-numbers.

; REMOVE AN ABBREVIATION:

; Examples:  (remove-abb '(1 2 3 4 5 6 7 8 9 10)).
;            (remove-abb (append x y))

; REMOVE-ABB causes the given term (which may be presented in ordinary
; user-level syntax) to no longer be abbreviated.

; REMOVE ALL ABBREVIATIONS:

; (clear-abbs)

; DISPLAY ALL ABBREVIATIONS:

; (show-abbs)

; Note:  Any term may be abbreviated.  In order to place a restriction
; on which terms may be abbreviated, change the function
; translate-and-filter below.  An example is given below the actual
; definition in which only quoted objects may be abbreviated.

; Technical point:  abbreviation works outside-in, so if term A is a
; subterm of term B and both have abbreviations, Nqthm will print the
; abbreviation of A when A is encountered outside-in.

(defvar untranslate-alist nil)

(defun translate-and-filter (form)
  ;; Could cause a soft error in selected cases, as in the following example.
  (translate form))

#|
(defun translate-and-filter (form)
  (let ((term (translate form)))
    (if (quotep term)
        term
      (er soft (form term)
          |Abbreviations| |are| |allowed| |only| |on| |explicit| |values| |,|
          |but| (!ppr form nil) |translates| |to| (!ppr term (quote |.|))))))
|#

(defun abb-fn (form val)
  (let ((term (translate-and-filter form)))
    (iterate for pair in untranslate-alist
             when (equal (cdr pair) val)
             do
             (cond
              ((equal (car pair) term)
               (er warning (form val) (!ppr form nil) |is| |already| |being|
                   |abbreviated| |as| (!ppr val (quote |.|)))
               (return-from abb-fn nil))
              (t (let (untranslate-alist)
                   (er warning ((x (car pair)) val form) |The| |abbreviation| (!ppr val nil)
                       |is| |already| |in| |use| |for| |the| |term| (!term x nil) |,|
                       |but| |we| |are| |now| |about| |to| |install| |it| |as| |an|
                       |abbreviation| |for| (!ppr form nil) |as| |well| |.|)))))
    (setq untranslate-alist
          (if (assoc-equal term untranslate-alist)
              (iterate for pair in untranslate-alist
                       collect
                       (if (equal (car pair) term)
                           (cons term val)
                         pair))
            (cons (cons term val) untranslate-alist))))
  t)

(defmacro abb (form val)
  `(abb-fn ',form ',val))

(defun clear-abbs ()
  (setq untranslate-alist nil))

(defun remove-abb-fn (form)
  (let ((term (translate form)))
    (if (assoc-equal term untranslate-alist)
        (progn (setq untranslate-alist
                     (iterate for pair in untranslate-alist
                              when (not (equal term (car pair)))
                              collect pair))
               t)
      (progn (er warning (form) (!ppr form nil)
                 |is| |not| |currently| |being| |abbreviated| |.|)
             nil))))

(defmacro remove-abb (form)
  `(remove-abb-fn ',form))

(defun show-abbs ()
  (let ((temp untranslate-alist)
        (untranslate-alist nil))
    (iterate for pair in temp
             do
             (prineval (pquote (progn cr (!term term nil) cr
                                      |  is| |abbreviated| |as| cr
                                      (!ppr val nil) cr))
                       (list (cons 'term (car pair))
                             (cons 'val (cdr pair)))
                       0 prove-file)))
  (values))

(DEFUN UNTRANSLATE (TERM &aux temp-pair)
  
;   Copies TERM, introducing abbreviations.  That is,
;   (TRANSLATE (UNTRANSLATE TERM)) = TERM.
  
  (COND ((setq temp-pair (assoc-equal term untranslate-alist))
         (cdr temp-pair))
        ((VARIABLEP TERM) TERM)
        ((QUOTEP TERM) (UNTRANSLATE-QUOTE (CADR TERM)))
        (T (CASE (FFN-SYMB TERM)
                 (IF
                     (COND ((CASE-FORMP TERM)
                            (CASE-FORM TERM))
                           ((COND-FORMP TERM)
                            (COND-FORM TERM))
                           (T (CONS 'IF
                                    (ITERATE FOR ARG IN (FARGS TERM)
                                             COLLECT (UNTRANSLATE ARG))))))
                 ((CAR CDR)
                  (ITERATE WITH L DO
                           (COND ((MATCH TERM (CAR &))
                                  (PUSH (QUOTE A) L)
                                  (SETQ TERM (FARGN TERM 1)))
                                 ((MATCH TERM (CDR &))
                                  (PUSH (QUOTE D) L)
                                  (SETQ TERM (FARGN TERM 1)))
                                 (T (RETURN
                                     `(,(PACK
                                         (CONS (QUOTE C)
                                               (NRECONC L (QUOTE (R)))))
                                       ,(UNTRANSLATE TERM)))))))
                 (CONS
                  (LET ((R (UNTRANSLATE (FARGN TERM 2))))
                    (COND ((NULL R) `(LIST ,(UNTRANSLATE (FARGN TERM 1))))
                          ((AND (CONSP R) (EQ (CAR R) (QUOTE LIST)))
                           `(LIST ,(UNTRANSLATE (FARGN TERM 1)) ,@(CDR R)))
                          (T `(CONS ,(UNTRANSLATE (FARGN TERM 1)) ,R)))))
                 ((AND OR PLUS TIMES)
                  (LET ((R (UNTRANSLATE (FARGN TERM 2))))
                    (COND ((AND (CONSP R) (EQ (CAR R) (FFN-SYMB TERM)))
                           `(,(CAR R) ,(UNTRANSLATE (FARGN TERM 1)) ,@(CDR R)))
                          (T `(,(FFN-SYMB TERM)
                               ,(UNTRANSLATE (FARGN TERM 1)) ,R)))))
                 (FOR
                  (COND
                   ((EQ *1*THM-MODE$ (QUOTE NQTHM))
                    (LET (V R C O B A)
                      (COND ((AND (MATCH TERM (FOR (LIST (QUOTE QUOTE) V)
                                                   R (LIST (QUOTE QUOTE) C)
                                                   (LIST (QUOTE QUOTE) O)
                                                   (LIST (QUOTE QUOTE) B) A))
                                  (TERMP V)
                                  (VARIABLEP V)
                                  (TERMP C)
                                  (MEMBER-EQ O FOR-OPS)
                                  (TERMP B)
                                  (EQUAL A (MAKE-ALIST V C B)))
                             (COND ((EQUAL C TRUE)
                                    (LIST (QUOTE FOR) V (QUOTE IN)
                                          (UNTRANSLATE R)
                                          O (UNTRANSLATE B)))
                                   (T (LIST (QUOTE FOR) V (QUOTE IN)
                                            (UNTRANSLATE R)
                                            (QUOTE WHEN) (UNTRANSLATE C)
                                            O (UNTRANSLATE B)))))
                            (T (CONS (QUOTE FOR)
                                     (ITERATE FOR ARG IN (FARGS TERM)
                                              COLLECT (UNTRANSLATE ARG)))))))
                   (T (CONS (FFN-SYMB TERM)
                            (ITERATE FOR ARG IN (FARGS TERM)
                                     COLLECT (UNTRANSLATE ARG))))))
                 (OTHERWISE (CONS (FFN-SYMB TERM)
                                  (ITERATE FOR ARG IN (FARGS TERM)
                                           COLLECT (UNTRANSLATE ARG))))))))
